unit VCLScriptControl;
(*******************************************************************

      TVCLScriptControl - VCL enhanced ScriptControl component

      Copyright (c) 1999 . by TAL
      All rights reserved

      This unit is free for non-commercial use

 *******************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, TypInfo, Controls,
  Forms, MSScriptControl_TLB, ActiveX, Graphics;

type
  LongWord = DWORD;
  PLongWord = PDWORD;

  EInvalidParamCount = class(Exception)
  end;

  EInvalidParamType = class(Exception)
  end;

  IQueryPersistent = interface
  ['{26F5B6E1-9DA5-11D3-BCAD-00902759A497}']
    function GetPersistent: TPersistent;
  end;

  //    D5  ( pceltFetched)
  IEnumVariant = interface(IUnknown)
    ['{00020404-0000-0000-C000-000000000046}']
    function Next(celt: LongWord; var rgvar : OleVariant;
      pceltFetched: PLongWord): HResult; stdcall;
    function Skip(celt: LongWord): HResult; stdcall;
    function Reset: HResult; stdcall;
    function Clone(out Enum: IEnumVariant): HResult; stdcall;
  end;

  TVCLScriptControl = class;

  TVCLProxy = class(TInterfacedObject, IDispatch, IQueryPersistent)
  private
    FOwner: TPersistent;
    FScriptControl: TVCLScriptControl;
    procedure DoCreateControl(AName, AClassName: WideString; WithEvents: Boolean);
    function SetVCLProperty(PropInfo: PPropInfo; Argument: TVariantArg): HRESULT;
    function GetVCLProperty(PropInfo: PPropInfo; dps: TDispParams;
      PDispIds: PDispIdList; var Value: OleVariant): HRESULT;
    { IDispatch }
    function GetTypeInfoCount(out Count: Integer): HResult; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
    { IQueryPersistent }
    function GetPersistent: TPersistent;
  protected
    function DoInvoke (DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var dps : TDispParams; pDispIds : PDispIdList;
      VarResult, ExcepInfo, ArgErr: Pointer): HResult; virtual;
  public
    constructor Create(AOwner: TPersistent; ScriptControl: TVCLScriptControl);
    destructor Destroy; override;
  end;

  TVCLEnumerator = class(TInterfacedObject, IEnumVariant)
  private
    FEnumPosition: Integer;
    FOwner: TPersistent;
    FScriptControl: TVCLScriptControl;
    { IEnumVariant }
    function Next(celt: LongWord; var rgvar : OleVariant;
      pceltFetched: PLongWord): HResult; stdcall;
    function Skip(celt: LongWord): HResult; stdcall;
    function Reset: HResult; stdcall;
    function Clone(out Enum: IEnumVariant): HResult; stdcall;
  public
    constructor Create(AOwner: TPersistent; AScriptControl: TVCLScriptControl);
  end;

  TVCLScriptControl = class(TScriptControl)
  private
    FProxyList: TList;
  protected
    { Protected declarations }
    function GetProxy(AOwner: TPersistent): IDispatch;
    procedure ProxyDestroyed(Address: Pointer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure RunProc(const Name: String);
    procedure RegisterClass(const Name: String; AClass: TPersistent);
  published
    { Published declarations }
    procedure OnChangeHandler(Sender: TObject);
    procedure OnClickHandler(Sender: TObject);
    procedure OnEnterHandler(Sender: TObject);
    procedure OnExitHandler(Sender: TObject);
    procedure OnTimerHandler(Sender: TObject);
  end;

procedure Register;

implementation

{ Utility routines }

function ValidType(Argument: TVariantArg; TypeId: Integer;
  RaiseException: Boolean): Boolean;
begin
 Result := Argument.vt = TypeId;
 if RaiseException and (not Result) then
   raise EInvalidParamType.Create('');
end;

function CheckArgCount(Count: Integer; Accepted: array of Integer;
  RaiseException: Boolean): Boolean;
var
  I: Integer;
begin
  Result := FALSE;
  for I := Low(Accepted) to High(Accepted) do begin
    Result := Accepted[I] = Count;
    if Result then
      Break;
  end;
  if RaiseException and (not Result) then
    raise EInvalidParamCount.Create('');
end;

function IntValue(Argument: TVariantArg): Integer;
var
 VT: Word;
 ByRef: Boolean;
begin
 VT := Argument.vt;
 ByRef := (VT and VT_BYREF) = VT_BYREF;
 if ByRef then begin
   VT := VT and (not VT_BYREF);
   case VT of
     VT_I1: Result := Argument.pbVal^;
     VT_I2: Result := Argument.piVal^;
     VT_VARIANT: Result := Argument.pvarVal^;
   else
     Result := Argument.plVal^;
   end;
 end else
 case VT of
   VT_I1: Result := Argument.bVal;
   VT_I2: Result := Argument.iVal;
 else
   Result := Argument.lVal;
 end;
end;

type
  PNamesArray = ^TNamesArray;
  TNamesArray = array[0..0] of PWideChar;
  PDispIdsArray = ^TDispIdsArray;
  TDispIdsArray = array[0..0] of Integer;
  TVariantList = array [0..0] of OleVariant;

const
  DISPID_CONTROLS    = 1;
  DISPID_COUNT       = 2;
  DISPID_ADD         = 3;
  DISPID_HASPROPERTY = 4;

{ TVCLProxy }

constructor TVCLProxy.Create(AOwner: TPersistent;
  ScriptControl: TVCLScriptControl);
begin
  inherited Create;
  FOwner := AOwner;
  FScriptControl := ScriptControl;
end;

function TVCLProxy.Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
  Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult;
var
  dps : TDispParams absolute Params;
  HasParams : boolean;
  pDispIds : PDispIdList;
  iDispIdsSize : integer;
  WS: WideString;
  I: Integer;
begin
  pDispIds := NIL;
  iDispIdsSize := 0;
  HasParams := (dps.cArgs > 0);
  if HasParams then
  begin
    iDispIdsSize := dps.cArgs * SizeOf(TDispId);
    GetMem(pDispIds, iDispIdsSize);
  end;
  try
    if HasParams then
      for I := 0 to dps.cArgs - 1 do
        pDispIds^[I] := dps.cArgs - 1 - I;
    try
      Result := DoInvoke(DispId, IID, LocaleID, Flags, dps, pDispIds, VarResult, ExcepInfo, ArgErr);
    except
      on E: EInvalidParamCount do Result := DISP_E_BADPARAMCOUNT;
      on E: EInvalidParamType do  Result := DISP_E_BADVARTYPE;
      on E: Exception do begin
        if Assigned(ExcepInfo) then begin
          FillChar(ExcepInfo^, SizeOf(TExcepInfo), 0);
          TExcepInfo(ExcepInfo^).wCode := 1001;
          TExcepInfo(ExcepInfo^).BStrSource := SysAllocString('TVCLProxy');
          WS := E.Message;
          TExcepInfo(ExcepInfo^).bstrDescription := SysAllocString(PWideChar(WS));
        end;
        Result := DISP_E_EXCEPTION;
      end;
    end;
  finally
    if HasParams then
      FreeMem(pDispIds, iDispIdsSize);
  end;
end;

function TVCLProxy.GetIDsOfNames(const IID: TGUID; Names: Pointer; NameCount,
  LocaleID: Integer; DispIDs: Pointer): HResult;
var
  S: String;
  Info: PPropInfo;
begin
  Result := S_OK;
  //     
  S := PNamesArray(Names)[0];
  // ,   VCL   - 
  Info := GetPropInfo(FOwner.ClassInfo, S);
  if Assigned(Info) then begin
    //  ,    DispId
    //   PropInfo
    PDispIdsArray(DispIds)[0] := Integer(Info);
  end else
  //   , ,    
  //     
  if CompareText(S, 'CONTROLS') = 0 then begin
    if (FOwner is TWinControl) then
      PDispIdsArray(DispIds)[0] := DISPID_CONTROLS
    else
      Result := DISP_E_UNKNOWNNAME;
  end
  else
  if CompareText(S, 'COUNT') = 0 then begin
    Result := S_OK;
    //      TCollection
    //  TStrings
    if (FOwner is TCollection) or (FOwner is TStrings)
       or (FOwner is TWinControl) then
      PDispIdsArray(DispIds)[0] := DISPID_COUNT
    else
      Result := DISP_E_UNKNOWNNAME;
  end
  else
  if CompareText(S, 'ADD') = 0 then begin
    Result := S_OK;
    //      TCollection
    //  TStrings
    if (FOwner is TCollection) or (FOwner is TStrings) or
       (FOwner is TWinControl) then
      PDispIdsArray(DispIds)[0] := DISPID_ADD
    else
      Result := DISP_E_UNKNOWNNAME;
  end
  else
  if CompareText(S, 'HASPROPERTY') = 0 then
    PDispIdsArray(DispIds)[0] := DISPID_HASPROPERTY
  else
    Result := DISP_E_UNKNOWNNAME;
end;

function TVCLProxy.DoInvoke(DispID: Integer; const IID: TGUID;
  LocaleID: Integer; Flags: Word; var dps: TDispParams;
  pDispIds: PDispIdList; VarResult, ExcepInfo, ArgErr: Pointer): HResult;

 function _ValidType(Index, TypeId: Integer; RaiseException: Boolean): Boolean;
 //     Index   
 begin
   Result := ValidType(dps.rgvarg^[pDispIds^[Index]], TypeId, RaiseException);
 end;

 function _IntValue(Index: Integer): Integer;
 //        Index
 begin
   Result := IntValue(dps.rgvarg^[pDispIds^[Index]]);
 end;

var
  S: String;
  Put: Boolean;
  I: Integer;
  P: TPersistent;
  B: Boolean;
  OutValue: OleVariant;
begin
  Result := S_OK;
  case DispId of
   DISPID_NEWENUM: begin
        //     IEnumVariant  ForEach
        //  ,   
        OleVariant(VarResult^) := TVCLEnumerator.Create(FOwner, FScriptControl) as IEnumVariant;
      end;
   DISPID_CONTROLS: begin
        //   Controls
        with FOwner as TWinControl do
        begin
          //  
          CheckArgCount(dps.cArgs, [1], TRUE);
          P := NIL;
          if _ValidType(0, VT_BSTR, FALSE) then begin
            //   -  -   
            //   
            S := dps.rgvarg^[pDispIds^[0]].bstrVal;
            for I := 0 to Pred(ControlCount) do
              if CompareText(S, Controls[I].Name) = 0 then begin
                P := Controls[I];
                Break;
              end;
          end else begin
            //  -  - ,    
            I := _IntValue(0);
            P := Controls[I];
          end;
          if not Assigned(P) then
            //   
            raise EInvalidParamType.Create('');
          //   IDispatch   
          OleVariant(VarResult^) := FScriptControl.GetProxy(P);
        end;
      end;
   DISPID_COUNT: begin
        //   Count
        // ,    
        CheckArgCount(dps.cArgs, [0], TRUE);
        if FOwner is TWinControl then
          //    
          OleVariant(VarResult^) := (FOwner as TWinControl).ControlCount;
        if FOwner is TCollection then
          //    
          OleVariant(VarResult^) := TCollection(FOwner).Count
        else
        if FOwner is TStrings then
          //   
          OleVariant(VarResult^) := TStrings(FOwner).Count;
      end;
   DISPID_ADD: begin
        //   Add
        if FOwner is TWinControl then begin
          //   
          CheckArgCount(dps.cArgs, [2,3], TRUE);
          //    
          _ValidType(0, VT_BSTR, TRUE);
          _ValidType(1, VT_BSTR, TRUE);
          //   - ,     -
          //  FALSE
          if (dps.cArgs = 3) and _ValidType(2, VT_BOOL, TRUE) then
            B := dps.rgvarg^[pDispIds^[0]].vbool
          else
            B := FALSE;
          //     
          DoCreateControl(dps.rgvarg^[pDispIds^[0]].bstrVal,
            dps.rgvarg^[pDispIds^[1]].bstrVal, B)
        end
        else
        if FOwner is TCollection then begin
          //  
          P := TCollection(FOwner).Add;
          //     IDispatch
          OleVariant(varResult^) := FScriptControl.GetProxy(P);
        end else
        if FOwner is TStrings then begin
          //   
          CheckArgCount(dps.cArgs, [1,2], TRUE);
          // ,   - 
          _ValidType(0, VT_BSTR, TRUE);
          if dps.cArgs = 2 then
            //   -   
            I := _IntValue(1)
          else
            //    -   
            I := TStrings(FOwner).Count;
          //  
          TStrings(FOwner).Insert(I, dps.rgvarg^[pDispIds^[0]].bstrVal);
        end;
      end;
   DISPID_HASPROPERTY: begin
        //   HasProperty
        //   
        CheckArgCount(dps.cArgs, [1], TRUE);
        //   
        _ValidType(0, VT_BSTR, TRUE);
        S := dps.rgvarg^[pDispIds^[0]].bstrVal;
        //  True,   
        OleVariant(varResult^) := Assigned(GetPropInfo(FOwner.ClassInfo, S));
      end;
  else
    //    ,   
    //  Flags,    
    //  
    Put := (Flags and DISPATCH_PROPERTYPUT) <> 0;
    if Put then begin
      //  
      //   
      CheckArgCount(dps.cArgs, [1], TRUE);
      //   
      Result := SetVCLProperty(PPropInfo(DispId), dps.rgvarg^[pDispIds^[0]])
    end else begin
      //  
      if DispId = 0 then begin
        // DispId = 0 -    
        //   IDispatch
        OleVariant(VarResult^) := Self as IDispatch;//IDispatch(TVCLProxy.Create(FOwner, FScriptControl));
        Result := S_OK;
        Exit;
      end;
      //   
      Result := GetVCLProperty(PPropInfo(DispId), dps, pDispIds, OutValue);
      if Result = S_OK then
        //   -  
        OleVariant(VarResult^) := OutValue;
    end;
  end;
end;

function TVCLProxy.GetTypeInfo(Index, LocaleID: Integer;
  out TypeInfo): HResult;
begin
  Pointer(TypeInfo) := NIL;
  Result := E_NOTIMPL;
end;

function TVCLProxy.GetTypeInfoCount(out Count: Integer): HResult;
begin
  Count := 0;
  Result := E_NOTIMPL;
end;

procedure TVCLProxy.DoCreateControl(AName, AClassName: WideString; WithEvents: Boolean);

  procedure SetHandler(Control: TPersistent; Owner:TObject; Name: String);
  //     Name   
  //   Name + 'Handler'
  var
    Method: TMethod;
    PropInfo: PPropInfo;
  begin
    //   RTTI
    PropInfo := GetPropInfo(Control.ClassInfo, Name);
    if Assigned(PropInfo) then begin
      //   
      Method.Code := FScriptControl.MethodAddress(Name + 'Handler');
      if Assigned(Method.Code) then begin
        //  
        Method.Data := FScriptControl;
        //  
        SetMethodProp(Control, PropInfo, Method);
      end;
    end;
  end;

var
  ThisClass: TControlClass;
  C: TComponent;
  NewOwner: TCustomForm;
begin
  //   Owner  
  if not (FOwner is TCustomForm) then
    NewOwner := GetParentForm(FOwner as TControl)
  else
    NewOwner := FOwner as TCustomForm;
  //    
  ThisClass := TControlClass(GetClass(AClassName));
  //  
  C := ThisClass.Create(NewOwner);
  //  
  C.Name := AName;
  if C is TControl then
    //   Parent
    TControl(C).Parent := FOwner as TWinControl;
  if WithEvents then begin
    //  
    SetHandler(C, NewOwner, 'OnClick');
    SetHandler(C, NewOwner, 'OnChange');
    SetHandler(C, NewOwner, 'OnEnter');
    SetHandler(C, NewOwner, 'OnExit');
    SetHandler(C, NewOwner, 'OnTimer');
  end;
  //     IDispatch   
  //    TScriptControl
  FScriptControl.RegisterClass(AName, C);
  //TVCLProxy.Create(C, FScriptControl) as IDispatch, False);
end;

function TVCLProxy.SetVCLProperty(PropInfo: PPropInfo;
  Argument: TVariantArg): HResult;
var
  I, J, K, CommaPos: Integer;
  GoodToken: Boolean;
  S, S1: String;
  DT: TDateTime;
  ST: TSystemTime;
  IP: IQueryPersistent;
  Data, TypeData: PTypeData;
  TypeInfo: PTypeInfo;
begin
  Result := S_OK;
  case PropInfo^.PropType^.Kind of
    tkChar, tkString, tkLString, tkWChar, tkWString: //  
      begin
        //   
        ValidType(Argument, VT_BSTR, TRUE);
        //   
        SetStrProp(FOwner, PropInfo, Argument.bstrVal);
      end;
    tkInteger:  //  
      begin
        //     TCursor, TColor
        //       
        //    
        if (UpperCase(PropInfo^.PropType^.Name) = 'TCURSOR') and
          (Argument.vt = VT_BSTR) then begin
          if not IdentToCursor(Argument.bstrVal, I) then begin
            Result := DISP_E_BADVARTYPE;
            Exit;
          end;
        end else
        if (UpperCase(PropInfo^.PropType^.Name) = 'TCOLOR') and
          (Argument.vt = VT_BSTR) then begin
          if not IdentToColor(Argument.bstrVal, I) then begin
            Result := DISP_E_BADVARTYPE;
            Exit;
          end;
        end else
          //  
          I := IntValue(Argument);
        //  
        SetOrdProp(FOwner, PropInfo, I);
      end;
    tkEnumeration:
      //  
      begin
        //    Boolean -    VBScript 
        //   
        if CompareText(PropInfo^.PropType^.Name, 'BOOLEAN') = 0 then begin
          //    
          ValidType(Argument, VT_BOOL, TRUE);
          //   Boolean -     
          SetOrdProp(FOwner, PropInfo, Integer(Argument.vBool));
        end else begin
          //  -       
          //    
          ValidType(Argument, VT_BSTR, TRUE);
          //  
          S := Trim(Argument.bstrVal);
          //   Integer
          I := GetEnumValue(PropInfo^.PropType^, S);
          //   -  
          if I >= 0 then
            SetOrdProp(FOwner, PropInfo, I)
          else
            raise EInvalidParamType.Create('');
        end;
      end;
    tkClass:
      //  
      begin
        //    -    IDispatch
        ValidType(Argument, VT_DISPATCH, TRUE);
        if Assigned(Argument.dispVal) then begin
          //   
          //   IQueryPersistent
          IP := IDispatch(Argument.dispVal) as IQueryPersistent;
          //    ,  
          //  
          I := Integer(IP.GetPersistent);
        end else
          //  -  
          I := 0;
        //  
        SetOrdProp(FOwner, PropInfo, I);
      end;
   tkFloat:
      //    
      begin
        if (PropInfo^.PropType^ = System.TypeInfo(TDateTime)) or
           (PropInfo^.PropType^ = System.TypeInfo(TDate)) then begin
          //    
          if Argument.vt = VT_BSTR then begin
            DT := StrToDate(Argument.bstrVal);
          end else begin
            ValidType(Argument, VT_DATE, TRUE);
            if VariantTimeToSystemTime(Argument.date, ST) <> 0 then
              DT := SystemTimeToDateTime(ST)
            else begin
              Result := DISP_E_BADVARTYPE;
              Exit;
            end;
          end;
          SetFloatProp(FOwner, PropInfo, DT);
        end else begin
          //    
          ValidType(Argument, VT_R8, TRUE);
          //  }
          SetFloatProp(FOwner, PropInfo, Argument.dblVal);
        end;
      end;
   tkSet:
      // 
      begin
        //   ,    
        ValidType(Argument, VT_BSTR, TRUE);
        //  
        S := Trim(Argument.bstrVal);
        //   RTTI
        Data := GetTypeData(PropInfo^.PropType^);
        TypeInfo := Data^.CompType^;
        TypeData := GetTypeData(TypeInfo);
        I := 0;
        while Length(S) > 0 do begin
          //   ,   
          //  
          CommaPos := Pos(',', S);
          if CommaPos = 0 then
            CommaPos := Length(S) + 1;
          S1 := Trim(System.Copy(S, 1, CommaPos - 1));
          System.Delete(S, 1, CommaPos);
          if Length(S1) > 0 then begin
            // ,     
            //  
            K := 1;
            GoodToken := FALSE;
            for J := TypeData^.MinValue to TypeData^.MaxValue do begin
              if CompareText(S1, GetEnumName(TypeInfo , J)) = 0 then begin
                //  ,    
                I := I or K;
                GoodToken := TRUE;
              end;
              K := K shl 1;
            end;
            if not GoodToken then begin
              //   
              Result := DISP_E_BADVARTYPE;
              Exit;
            end;
          end;
        end;
        //   
        SetOrdProp(FOwner, PropInfo, I);
      end;
    tkVariant:
      // 
      begin
        //    
        ValidType(Argument, VT_VARIANT, TRUE);
        //  
        SetVariantProp(FOwner, PropInfo, Argument.pvarVal^);
      end;
   else
     //    OLE  
     Result := DISP_E_MEMBERNOTFOUND;
  end;
end;

function TVCLProxy.GetVCLProperty(PropInfo: PPropInfo; dps: TDispParams;
  PDispIds: PDispIdList; var Value: OleVariant): HResult;
var
  I, J, K: Integer;
  S: String;
  P, P1: TPersistent;
  Data: PTypeData;
  DT: TDateTime;
  TypeInfo: PTypeInfo;
begin
  Result := S_OK;
  case PropInfo^.PropType^.Kind of
    tkString, tkLString, tkWChar, tkWString:
      //  
      Value := GetStrProp(FOwner, PropInfo);
    tkChar, tkInteger:
      //  
      Value := GetOrdProp(FOwner, PropInfo);
    tkEnumeration:
      //  
      begin
        // ,  Boolean  
        if CompareText(PropInfo^.PropType^.Name, 'BOOLEAN') = 0 then begin
          //   Boolean
          Value := Boolean(GetOrdProp(FOwner, PropInfo));
        end else begin
          //  -   
          I := GetOrdProp(FOwner, PropInfo);
          Value := GetEnumName(PropInfo^.PropType^, I);
        end;
      end;
    tkClass:
      // 
      begin
        //   
        P := TPersistent(GetOrdProp(FOwner, PropInfo));
        if Assigned(P) and (P is TCollection) and (dps.cArgs = 1) then begin
          //      ( )
          if ValidType(dps.rgvarg^[pDispIds^[0]], VT_BSTR, FALSE) then begin
            //  ,    
            // DisplayName
            S := dps.rgvarg^[pDispIds^[0]].bstrVal;
            P1 := NIL;
            for I := 0 to Pred(TCollection(P).Count) do
              if CompareText(S, TCollection(P).Items[I].DisplayName) = 0 then begin
                P1 := TCollection(P).Items[I];
                Break;
              end;
            if Assigned(P1) then
              //  -   IDispatch
              Value := FScriptControl.GetProxy(P1)
               //IDispatch(TVCLProxy.Create(P1, FScriptControl))
            else
              //  
              Result := DISP_E_MEMBERNOTFOUND;
          end else begin
            //  ,    
            I := IntValue(dps.rgvarg^[pDispIds^[0]]);
            if (I >= 0) and (I < TCollection(P).Count) then begin
              P := TCollection(P).Items[I];
              Value := FScriptControl.GetProxy(P);
                //IDispatch(TVCLProxy.Create(P, FScriptControl));
            end else
              Result := DISP_E_MEMBERNOTFOUND;
          end;
        end
        else
        if Assigned(P) and (P is TStrings) and (dps.cArgs = 1) then begin
          //    Strings   ( )
          if ValidType(dps.rgvarg^[pDispIds^[0]], VT_BSTR, FALSE) then begin
            //   -    Values
            S := dps.rgvarg^[pDispIds^[0]].bstrVal;
            Value := TStrings(P).Values[S];
          end else begin
            //  ,    
            I := IntValue(dps.rgvarg^[pDispIds^[0]]);
            if (I >= 0) and (I < TStrings(P).Count) then
              Value := TStrings(P)[I]
            else
              Result := DISP_E_MEMBERNOTFOUND;
          end;
        end
        else
          //  ,   IDispatch 
          if Assigned(P) then
            Value := FScriptControl.GetProxy(P)
              //IDispatch(TVCLProxy.Create(P, FScriptControl))
          else
            //  Unassigned,   = NIL
            Value := Unassigned;
      end;
    tkFloat:
      //    
      begin
        if (PropInfo^.PropType^ = System.TypeInfo(TDateTime)) or
           (PropInfo^.PropType^ = System.TypeInfo(TDate)) then begin
          DT := GetFloatProp(FOwner, PropInfo);
          Value := DT;
        end else
          Value := GetFloatProp(FOwner, PropInfo);
      end;
    tkSet:
      // 
      begin
        //    ( )
        I := GetOrdProp(FOwner, PropInfo);
        //   RTTI
        Data := GetTypeData(PropInfo^.PropType^);
        TypeInfo := Data^.CompType^;
        //     
        S := '';
        if I <> 0 then begin
          for K := 0 to 31 do begin
            J := 1 shl K;
            if (J and I) = J then
              S := S + GetEnumName(TypeInfo, K) + ',';
          end;
          System.Delete(S, Length(S), 1);
        end;
        Value := S;
      end;
    tkVariant:
      // 
      Value := GetVariantProp(FOwner, PropInfo);
  else
    //    
    Result := DISP_E_MEMBERNOTFOUND;
  end;
end;

function TVCLProxy.GetPersistent: TPersistent;
begin
  //  IQueryPersistent
  //    
  Result := FOwner;
end;

destructor TVCLProxy.Destroy;
begin
  //      TVCLProxy
  FScriptControl.ProxyDestroyed(Self);
  inherited;
end;

{ TVCLEnumerator }

constructor TVCLEnumerator.Create(AOwner: TPersistent;
  AScriptControl: TVCLScriptControl);
begin
  inherited Create;
  FOwner := AOwner;
  FScriptControl := AScriptControl;
  FEnumPosition := 0;
end;

function TVCLEnumerator.Clone(out Enum: IEnumVariant): HResult;
var
  NewEnum: TVCLEnumerator;
begin
  NewEnum := TVCLEnumerator.Create(FOwner, FScriptControl);
  NewEnum.FEnumPosition := FEnumPosition;
  Enum := NewEnum as IEnumVariant;
  Result := S_OK;
end;

function TVCLEnumerator.Next(celt: LongWord; var rgvar: OleVariant;
  pceltFetched: PLongWord): HResult;
var
  I: Cardinal;
begin
  Result := S_OK;
  I := 0;
  if FOwner is TWinControl then begin
    with TWinControl(FOwner) do begin
      while (FEnumPosition < ControlCount) and (I < celt) do begin
        TVariantList(rgvar)[I] :=
          FScriptControl.GetProxy(Controls[FEnumPosition]);
//          TVCLProxy.Create(Controls[FEnumPosition], FScriptControl) as IDispatch;
        Inc(I);
        Inc(FEnumPosition);
      end;
    end;
  end
  else
  if FOwner is TCollection then begin
    with TCollection(FOwner) do begin
      while (FEnumPosition < Count) and (I < celt) do begin
        TVariantList(rgvar)[I] :=
          FScriptControl.GetProxy(Items[FEnumPosition]);
//          TVCLProxy.Create(Items[FEnumPosition], FScriptControl) as IDispatch;
        Inc(I);
        Inc(FEnumPosition);
      end;
    end;
  end
  else
  if FOwner is TStrings then begin
    with TStrings(FOwner) do begin
      while (FEnumPosition < Count) and (I < celt) do begin
        TVariantList(rgvar)[I] := TStrings(FOwner)[FEnumPosition];
        Inc(I);
        Inc(FEnumPosition);
      end;
    end;
  end else
    Result := S_FALSE;
  if I <> celt then
    Result := S_FALSE;
  if Assigned(pceltFetched) then
    pceltFetched^ := I;
end;

function TVCLEnumerator.Reset: HResult;
begin
  FEnumPosition := 0;
  Result := S_OK;
end;

function TVCLEnumerator.Skip(celt: LongWord): HResult;
var
  Total: Integer;
begin
  Result := S_FALSE;
  if FOwner is TWinControl then
    Total := TWinControl(FOwner).ControlCount
  else
  if FOwner is TCollection then
    Total := TCollection(FOwner).Count
  else
  if FOwner is TStrings then
    Total := TStrings(FOwner).Count
  else
    Exit;
  {$WARNINGS OFF}
  if FEnumPosition + celt <= Total then begin
  {$WARNINGS ON}
    Result := S_OK;
    Inc(FEnumPosition, celt)
  end;
end;

{ TVCLScriptControl }

constructor TVCLScriptControl.Create(AOwner: TComponent);
begin
  inherited;
  FProxyList := TList.Create;
end;

destructor TVCLScriptControl.Destroy;
begin
  FProxyList.Free;
  FProxyList := NIL;
  inherited;
end;

function TVCLScriptControl.GetProxy(AOwner: TPersistent): IDispatch;
var
  I: Integer;
  QP: IQueryPersistent;
begin
  Result := NIL;
  with FProxyList do
    for I := 0 to Pred(Count) do begin
      if TObject(FProxyList[I]).GetInterface(IQueryPersistent, QP) and
         (QP.GetPersistent = AOwner) then begin
        Result := TVCLProxy(FProxyList[I]) as IDispatch;
        Exit;
      end;
    end;
  I := FProxyList.Add(TVCLProxy.Create(AOwner, Self));
  Result := TVCLProxy(FProxyList[I]) as IDispatch;
end;

procedure TVCLScriptControl.ProxyDestroyed(Address: Pointer);
begin
  if Assigned(FProxyList) then
    with FProxyList do
      Delete(IndexOf(Address));
end;

procedure TVCLScriptControl.RunProc(const Name: String);
var
  P: PSafeArray;
  VarBound: TVarArrayBound;
  Proc: IScriptProcedure;
begin
  try
    Proc := Procedures[Name]
  except
    Proc := NIL;
  end;
  if Assigned(Proc) then begin
    FillChar(VarBound, SizeOf(VarBound), 0);
    VarBound.ElementCount := 0;
    P := SafeArrayCreate(varVariant, 1, VarBound);
    try
      Run(Name, P);
    finally
      SafeArrayDestroy(P);
    end;
  end;
end;

procedure TVCLScriptControl.RegisterClass(const Name: String;
  AClass: TPersistent);
begin
  AddObject(Name, GetProxy(AClass), FALSE);
end;

procedure TVCLScriptControl.OnChangeHandler(Sender: TObject);
begin
  RunProc((Sender as TComponent).Name + '_' + 'OnChange');
end;

procedure TVCLScriptControl.OnClickHandler(Sender: TObject);
begin
  RunProc((Sender as TComponent).Name + '_' + 'OnClick');
end;

procedure TVCLScriptControl.OnEnterHandler(Sender: TObject);
begin
  RunProc((Sender as TComponent).Name + '_' + 'OnEnter');
end;

procedure TVCLScriptControl.OnExitHandler(Sender: TObject);
begin
  RunProc((Sender as TComponent).Name + '_' + 'OnExit');
end;

procedure TVCLScriptControl.OnTimerHandler(Sender: TObject);
begin
  RunProc((Sender as TComponent).Name + '_' + 'OnTimer');
end;

procedure Register;
begin
  RegisterComponents('ActiveX', [TVCLScriptControl]);
end;

initialization
  OleInitialize(NIL);

finalization
  OleUninitialize;

end.
